import copy
import numpy as np
from skimage.morphology import binary_dilation, remove_small_objects, disk

def cropImage(image, crop_bottom, crop_top, crop_left, crop_right):
    """
    Crop an image
    
    Parameters
    ----------
      image : jpg or png
          The image that is to be cropped
      crop_bottom, crop_top, crop_left, crop: int
          How many pixels to crop in each of these directions
        
    Returns
    ---------
      image: array
          The image that has been cropped
    """

    if crop_bottom == 0:
        image = image[crop_top:]
    else:
        image = image[crop_top:-crop_bottom]
    if crop_right == 0:
        image = image[:, crop_left:]
    else:
        image = image[:, crop_left:-crop_right]
    
    return image

def cropping_tube(image, crop_param, size_param, dilation_param):
    """
    Cropping tubes of an image
      
    Parameters
    ----------
      image: jpg, png
          The original tubed image that needs to be cropped
      crop_param: int
          Threshold for removing the dark edges i.e., tube ends in the image
      size_param: int
          Make sure features below this size (i.e. hydrides) are not included in cropping
      dilation_param:int
          Dilate the cropped boundary by a number of pixels
    
    Returns
    -------
      cropped image:
        The final cropped image
      cropped threshold:
        True/False array highlighting the cropped and not cropped regions
    """
    
    crop_threshold = image < crop_param
    crop_threshold = remove_small_objects(crop_threshold, size_param)

    crop_threshold = binary_dilation(crop_threshold, selem=disk(dilation_param))        
    cropped_image = copy.deepcopy(image)
    cropped_image[crop_threshold] = np.nan

    return cropped_image, crop_threshold



import numpy as np
from matplotlib_scalebar.scalebar import ScaleBar
from matplotlib import pyplot as plt

def addScaleBar(ax, scale, location='upper right'):
    """
    Add a scale bar to an axes.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Matplotlib axis to plot on 
    """

    if scale:
        scalebar = ScaleBar(scale, location=location)
        ax.add_artist(scalebar)
        plt.show()

def addArrows(ax, c='r', lenx=0.04, leny=0.06, flip=False):
    """
    Add coordinate definition arrows (radial and circumferential) to an axes.
    
    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Matplotlib axis to plot on 
    """

    startcoord = (0.02, 0.02);
    
    ax.annotate("", xy=(startcoord[0]-0.002, startcoord[1]), xytext=(startcoord[0]+lenx, startcoord[1]), 
        xycoords = 'axes fraction', c=c, arrowprops=dict(arrowstyle="<-", color=c, lw=2))
    ax.annotate("", xy=(startcoord[0], startcoord[1]-0.002), xytext=(startcoord[0], startcoord[1]+leny), 
        xycoords = 'axes fraction', c=c, arrowprops=dict(arrowstyle="<-", color=c, lw=2))
    if flip == False:
        ax.annotate("R", xy=(0.011, startcoord[1]+leny), xycoords = 'axes fraction', fontsize=14, fontweight='bold', c=c)
        ax.annotate("C", xy=(startcoord[0]+lenx, 0.01), xycoords = 'axes fraction', fontsize=14, fontweight='bold', c=c)
    if flip == True:
        ax.annotate("C", xy=(0.011, startcoord[1]+leny), xycoords = 'axes fraction', fontsize=14, fontweight='bold', c=c)
        ax.annotate("R", xy=(startcoord[0]+lenx, 0.01), xycoords = 'axes fraction', fontsize=14, fontweight='bold', c=c)

def plot(img, title, scale=None, location=None):
    """
    Plotting an imge.
    
    Parameters
    ----------
        img:
            Image data to be plotted
        title: str
            Title of plot.
        scale: float
            Scale in meters per pixel.
        location: str
            Location of scale bar i.e. 'lower right', 'upper left'
    """

    fig, ax = plt.subplots(figsize=(10,6))
    plt.imshow(img, cmap='gray')
    plt.title(title, fontsize=14)
    plt.axis('off');
    if scale is not None:
        addScaleBar(ax, scale=scale, location=location)
    addArrows(ax)

def plot_comparison(img1, title1, img2, title2, scale=None, location=None):
    """
    Plotting two images next to each other.
    
    Parameters
    ----------
    img1:
        Image data to be plotted on left.
    img2:
        Image data to be plotted on right.
    title1: str
        Title of left-hand plot.
    title2: str
        Title of right-hand plot.
    scale: float
        Scale in meters per pixel.
    location: str
        Location of scale bar i.e. 'lower right', 'upper left'
    """

    fig, (ax_a, ax_b) = plt.subplots(ncols=2, figsize=(14, 7), sharex=True, sharey=True)
    ax_a.imshow(img1, cmap=plt.cm.gray)
    ax_a.set_title(title1, fontsize=14)
    ax_a.axis('off')
    ax_b.imshow(img2, cmap=plt.cm.gray)
    ax_b.set_title(title2, fontsize=14)
    ax_b.axis('off')
    addArrows(ax_a)
    addArrows(ax_b)
    plt.tight_layout()
    if scale is not None:
        addScaleBar(ax_b, scale, location)

def plot_hist(arr):
    fig, ax = plt.subplots(figsize=(8,5))
    histogram = ax.hist(arr[~np.isnan(arr)].flatten(), bins=60, range=(0,2));
    ax.set_title('Image Histogram', fontsize=14);
    ax.set_xlabel('Gray value', fontsize=12); ax.set_ylabel('Frequency', fontsize=12);

    return histogram
    


import copy
import numpy as np
from scipy.ndimage import gaussian_filter
from skimage.morphology import remove_small_objects


def nan_gaussian(image, sigma):
    """
    Apply a gaussian filter to an array with nans.
    
    Parameters
    ----------
        image: 
            an array with nans
        sigma: int
            σ is the standard deviation of the Gaussian distribution
    
    Returns
    -------
        gauss
    """

    nan_msk = np.isnan(image)
    loss = np.zeros(image.shape)
    loss[nan_msk] = 1
    loss = gaussian_filter(loss, sigma=sigma, mode='constant', cval=1)
    gauss = image.copy()
    gauss[nan_msk] = 0
    gauss = gaussian_filter(gauss, sigma=sigma, mode='constant', cval=0)
    gauss[nan_msk] = np.nan
    gauss += loss * image

    return gauss

def minimize_grain_contrast(image, sigma):
    """
    Minimise grain contrast or uneven lighting by dividing the original image 
    by an image with a gaussian blur applied.
    
    Parameters
    ----------
        image: 
            Image to minimise grain contrast.
        sigma: float
            Sigma value for gaussian blur.
        
    Returns
    -------
        removedGrains:
            Output image.
    """

    gaussian_blur = nan_gaussian(image, sigma = sigma)
    removedGrains = image / gaussian_blur
    
    return(removedGrains)

def simple_threshold(image, crop_threshold, threshold, small_obj=None):
    """
    Threshold the image, where the hydrides are identified to be white and the matrix is black
    
    Parameters
    ----------
        removed_grains: arr
            image to threshold
        theshold: float
            threshold level
        small_grains: int
            size of features to be removed and not thresholded
        crop_threshold: array
            array of true false so that thresholding is only performed within the region that has been previously cropped
        
    Returns
    -------
        thres_disp:
            The thresholded image
    """

    thres = image < threshold
    if small_obj is not None:
        thres = remove_small_objects(thres, min_size=small_obj)


    thres_disp = copy.deepcopy(thres)
    thres_disp = np.array(thres_disp)*1.0
    thres_disp[crop_threshold] = np.nan
           
    return thres_disp




import numpy as np
from matplotlib import pyplot as plt
from skimage.morphology import skeletonize
from skan import Skeleton, summarize
import networkx as nx
import toolz as tz

def branch_classification(thres):
    """
    Predict the extent of branching
    
    Parameters
    ----------
        thres: array
            thresholded image to be analysed
    
    Returns
    -------
        skel: array
            skeletonised image
        is_main:
            help
        BLF: int/float
            branch length fraction
    """
    
    skeleton = skeletonize(thres)
    skel = Skeleton(skeleton, source_image=thres)
    summary = summarize(skel)
    
    is_main = np.zeros(summary.shape[0])
    us = summary['node-id-src']
    vs = summary['node-id-dst']
    ws = summary['branch-distance']

    edge2idx = {
        (u, v): i
        for i, (u, v) in enumerate(zip(us, vs))
    }



    edge2idx.update({
        (v, u): i
        for i, (u, v) in enumerate(zip(us, vs)) 
    })

    g = nx.Graph()

    g.add_weighted_edges_from(
        zip(us, vs, ws)
    )

    for conn in nx.connected_components(g):
        curr_val = 0
        curr_pair = None
        h = g.subgraph(conn) 
        p = dict(nx.all_pairs_dijkstra_path_length(h))
        for src in p:
            for dst in p[src]:
                val = p[src][dst]
                if (val is not None
                        and np.isfinite(val)
                        and val > curr_val):
                    curr_val = val 
                    curr_pair = (src, dst) 
        for i, j in tz.sliding_window(
            2,
            nx.shortest_path(
                h, source=curr_pair[0], target=curr_pair[1], weight='weight'
            )
        ):
            is_main[edge2idx[(i, j)]] = 1 
            
    summary['main'] = is_main

    #Branch Length Fraction

    total_length = np.sum(skeleton)
    trunk_length = 0
    for i in range(summary.shape[0]):
        if summary['main'][i]:
            trunk_length += summary['branch-distance'][i]

    branch_length = total_length - trunk_length
    BLF = branch_length/total_length

    return skel,is_main,BLF


import numpy as np
from skimage.color import rgb2gray
from skimage.io import imread

def image(image_path, transpose):
    """
    Import and Transpose an image.

    Parameters
    ----------
        image_path: str
            Path to the image, excluding extension, including name
        transpose: Boolean Value
            Whether the image needs to be rotated so that radial hydrides are in the vertical direction
            If True: Image will be transposed
            If False: Image will not be transposed

    Returns
    -------
        original_image
            The transposed/imported image to be analysed
    """

    original_image = imread(image_path)
    original_image = rgb2gray(original_image)

    if transpose == True:
        original_image = np.transpose(original_image)

    return original_image



import numpy as np
from scipy import ndimage
from skimage.graph import route_through_array
from skimage.transform import rescale
from scipy.ndimage.morphology import binary_dilation

def det_crack_path(thres, crop_threshold, num_runs, kernel_size):
    """
    Determine Possible Crack Paths in the micrograph
    
    Parameters
    ----------
        thres: array
            thesholded image to look at
        crop_threshold: array
            calculated during thresholding, array of true and false values
        num_runs int
            number of crack paths to determine
    
    Returns
    -------
        edist: array
            min euclidean distace from the hydride to the matrix
        path_list:
            list of possible crack paths
        cost_list: list
            list of cost values for each crack path
    """
  
    # Use distance away from a hydride as a path to route through
    edist = ndimage.morphology.distance_transform_edt(thres==0)
    edist = rescale(edist,(1, 1))

    # Add a row of zeros on the top and bottom and set cost=0 outside tube
    edist[0,:] = 0; edist[-1,:] = 0; 
    edist = np.where(crop_threshold==True, 0, edist)

    # Make a empty list to store paths and costs
    path_list = []; cost_list = [];

    for run in np.arange(num_runs):
        # Coordinates and cost corresponding to path
        path, cost = route_through_array(edist,[0,0],[-1,-1])
        path = np.array(path)
      
        # Boolean array based on coordinates, True is on path
        path_array = np.zeros(np.shape(edist), dtype=bool)
        for coord in path:
            path_array[coord[0], coord[1]] = True

        # Take away points outside of crop, make coord array and append to list
        path_array_cropped = np.logical_and(path_array, ~crop_threshold)
        path_coords = np.transpose(np.nonzero(path_array_cropped==True))
        path_list.append(np.array(path_coords))
        cost_list.append(cost)

        # Filtering path based on kernel size, so that the next run will take a different path
        filter_array = binary_dilation(path_array_cropped, iterations=kernel_size)
        edist = np.where(filter_array==True, np.inf, edist)
        edist = np.where(crop_threshold==True, 0, edist)

    edist = ndimage.morphology.distance_transform_edt(thres == 0)
    edist = rescale(edist,(1,1))


    return edist, path_list, cost_list



import numpy as np
from scipy import ndimage
from skimage.transform import hough_line, hough_line_peaks
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle
from .plot_functions import addScaleBar, addArrows

def hough_rad(image, num_peaks, min_distance=5, min_angle=5, val=0.25, scale=None, location=None):
    """
    Perform Hough Line Transfom to determine radial hydride fraction
    
    Parameters
    ----------
        image: array
            input thresholded hydride image
        num_peaks: int/float
            number of measured peaks in one rectangle 
        min_distance: int/float
            minimum distance separating two parallel lines it seems that a value of 5 is good
        min_angle: int/float
            minimum angle separating two lines it seems that a value of 5 is good
        val: float
            val is a number < 1 where, only hydrides that are at least val times the length of the 
            longest hydride are measured. This helps to reduce noise becuase only hydrides that are significant 
            in size are included in the calculation. The default value for this is 0.25, if you have much smaller hydride branches that you want to pick up
            this value can be reduced, but remember the noise increases as well.
        scale: float
            Scale in meters per pixel.
        location: str
            Location of scale bar i.e. 'lower right', 'upper left'

    Returns
    -------
        Radial:
            Fraction of radial hydrides calculated from Hough Transform
        Circumferential:
            Fraction of circumferential hydrides calculated from Hough Transform
        angle_list: List
            List of angles generated from the hough transform
    """

    fig, axes = plt.subplots(ncols=2, figsize=(14, 7), sharex=True, sharey=True)
    ax = axes.ravel()

    # Plotting
    ax[0].imshow(image, cmap='gray')
    ax[0].set_axis_off()
    ax[0].set_title('Thresholded image', fontsize=14)
    ax[1].imshow(image, cmap='gray')
    ax[1].set_axis_off()
    ax[1].set_title('Hough Transform', fontsize=14)

    # Label image
    label, num_features = ndimage.label(image > 0.1)
    slices = ndimage.find_objects(label)

    # Loop over each slice
    len_list = np.zeros([0]); angle_list = np.zeros([0]); d_list = np.zeros([0]);
    for feature in np.arange(num_features):
        h, theta, d = hough_line(label[slices[feature]], theta=np.linspace(-np.pi/2 , np.pi/2 , 90))
        threshold = val*np.amax(h)
        h_peak, angles, d_peak = hough_line_peaks(h, theta, d, threshold=threshold, 
            num_peaks=num_peaks, min_distance=min_distance, min_angle=min_angle)
        angle_list = np.append(angle_list,angles)
        len_list = np.append(len_list,h_peak)
        d_list = np.append(d_list, d_peak)

        # Draw bounding box
        x0_box = np.min([slices[feature][1].stop, slices[feature][1].start])
        y0_box = np.min([slices[feature][0].stop, slices[feature][0].start])
        x1_box = np.max([slices[feature][1].stop, slices[feature][1].start])
        y1_box = np.max([slices[feature][0].stop, slices[feature][0].start])
        rect = Rectangle((x0_box, y0_box), x1_box-x0_box, y1_box-y0_box, angle=0.0, ec='r', fill=False)
        ax[1].add_artist(rect)
        origin=np.array((0, np.abs(x1_box-x0_box)))
        for _, angle, dist in zip(h_peak, angles, d_peak):
            y0b, y1b = (dist - np.array((0, x1_box-x0_box)) * np.cos(angle)) / np.sin(angle)
            y0_line = y0b+y0_box; y1_line = y1b+y0_box
            x0_line = x0_box; x1_line = x1_box
            m = (y1_line-y0_line)/(x1_line-x0_line)

            # Fix lines which go over the edges of bounding boxes
            if y0_line < y0_box: x0_line = ((y0_box - y1_line) / m) + x1_line; y0_line = y0_box;
            if y0_line > y1_box: x0_line = ((y1_box - y1_line) / m) + x1_line; y0_line = y1_box;
            if y1_line < y0_box: x1_line = ((y0_box - y1_line) / m) + x1_line; y1_line = y0_box;
            if y1_line > y1_box: x1_line = ((y1_box - y1_line) / m) + x1_line; y1_line = y1_box;

            ax[1].plot(np.array((x0_line, x1_line)), (y0_line, y1_line), '-g')

    print('Number of detected angles: {0}'.format(len(len_list)))
    ax[1].set_xlim(0,image.shape[1]); ax[0].set_ylim(0,image.shape[0]);

    if scale is not None:
        addScaleBar(ax[1], scale, location)
    addArrows(ax[0])
    addArrows(ax[1])

    plt.tight_layout()

    return angle_list,len_list


def RHF_no_weighting_factor(angle_list, len_list):
    """
    Calculate the Radial Hydride Fraction without any weighting factor

    Parameters
    ----------
    angle_list:list
        calculated from the Hough line transform
    len_list: list
        List of lengths generated from the hogh line transform

    Returns
    ----------
    radial: float 
        fraction of radial hydrides
    circumferential: float
        fraction of circumferential hydrides
    """
    
    radial = np.sum(np.where(np.logical_and(-np.pi / 4 <= angle_list, angle_list < np.pi / 4), len_list, 0))
    circumferential = np.sum(np.where(np.logical_and(-np.pi / 4 <= angle_list, angle_list < np.pi / 4), 0, len_list))
    
    radial = radial / (radial + circumferential)
    circumferential = circumferential / (radial + circumferential)
    
    return radial, circumferential 


def weighted_RHF_calculation(angle_list, len_list):
    """
    Weighted Radial Hydride Fraction Calculation
    
    Parameters
    ----------
    angle_list: list
        List of angles generated from the hogh line transform
    len_list: list
        List of lengths generated from the hogh line transform
   
    Returns
    -------
      RHF: float
          Weighted radial hydride fraction
    """

    deg_angle_list = np.rad2deg(angle_list)
    fi = []

    for k in deg_angle_list:
        if k >0 and k<=30: x=1
        elif k>30 and k<=50: x=0.5
        elif k>50 and k<=90: x=0
        elif k>-30 and k<=0: x=1
        elif k>-50 and k<=-30: x=0.5
        elif k>-90 and k<=-50: x=0

        fi.append(x)

    #The next step is to do the summation
    SumOfLixFi = sum(len_list * np.array(fi))
    SumOfLi = sum(len_list)

    RHF = SumOfLixFi / SumOfLi

    return RHF




from setuptools import setup, find_packages


def get_long_description():
    readme_file = 'README.md'
    with open(readme_file, encoding='utf-8') as handle:
        contents = handle.read()

    return contents


setup(
    name='RHaP',
    version='0.2',
    author='Mia Maric, Rhys Thomas, Michael D. Atkinson',
    author_email='mia.maric@manchester.ac.uk',
    description='A python library to analyse hydride morphology present in light optical and scanning electron micrographs.',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    keywords='micrograph, zirconium, hydride',
    project_urls={
        'GitHub': 'https://github.com/miamaric12/Radial_Hydride_Code'
    },
    classifiers=[
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Framework :: IPython',
        'Framework :: Jupyter',
        'Framework :: Matplotlib'
    ],
    packages=find_packages(),
    package_data={'RHaP': ['data/example.bmp']},
    python_requires='>=3.5',
    install_requires=[
        'scipy',
        'numpy',
        'matplotlib>=3.0.0',
        'scikit-image',
        'numba',
        'skan',
        'toolz',
        'matplotlib_scalebar',
        'networkx',
        'jupyter'
    ],
)
